from __future__ import annotations

import re
import sys
import shutil
import os.path
from typing import Any
from pathlib import Path

import tricot
import tricot.utils
from tricot.command import Command


this = sys.modules[__name__]
this.validators = {}


def register_validator(validator_name: str, validator_class: type) -> None:
    '''
    Registers a validator class under the specified name. This function needs to
    be called to make validators available within of .yml files. Some default
    validator classes are registerd at the end of this file. Others can be manually
    added by users that use tricot as a library.

    Parameters:
        validator_name      Name of the validator to use in .yml files
        validator_class     Reference to the corresponding validator class

    Returns:
        None
    '''
    this.validators[validator_name] = validator_class


def get_validator(path: Path, validator_name: str, param: Any, variables: dict[str:str]) -> Validator:
    '''
    Searches for the specified validator name within the registred validators and
    uses the corresponding class to create an instance for the validator. This instance
    is returned by the function. If the specified validator name is not found, the
    function raises a ValidatorError.

    Parameters:
        path                Path to the .yml file that contains the validator
        validator_name      Name of the requested validator
        param               Params to initialize the validator with
        variables           Variables to initialize the validator with
    '''
    val_class = this.validators.get(validator_name)

    if val_class is None:
        raise ValidatorError(path, f"Unable to find specified validator '{validator_name}'.")

    return val_class(path, validator_name, param, variables)


def get_validator_list() -> list[str]:
    '''
    Returns a list of currently registered validator names.

    Parameters:
        None

    Returns:
        validator_list      List of registred validators
    '''
    keys = this.validators.keys()
    return list(keys)


class ValidationException(Exception):
    '''
    ValidationExceptions are raised by Validators when their validation procedure
    failed. Custom validators should always raise this exception to indicate a
    failure.
    '''


class ValidatorError(Exception):
    '''
    ValidatorErrors are raised when a Validator was specified with incorrect
    parameters. The corresponding Exception just contains a message describing
    the problem.
    '''
    def __init__(self, path: Path, message: str) -> None:
        '''
        Just add one additional attribute that is the path to the config file where the validator is
        defined in.
        '''
        self.path = str(path.resolve())
        super().__init__(message)


class Validator:
    '''
    The Validator class is basically an abstract class with the only purpose of
    being extended by other classes. All possible validation options implemented in
    tricot are based on this superclass that contains some shared methods that are
    useful for all validators.

    Within test.yaml files, Validators can be defined in different ways:

        * example-validator: This is a value
        * example-validator2:
            param1: value1
            param2: value2
            ...
        * example-validator3:
            - value1
            - value2
            ...

    Which form to choose depends on the corresponding Validator requirements. When
    choosing the first version, the 'param' attribute of the Validator just contains
    the value specified in the yaml file. When choosing the second or third form, the
    'param' attribute contains a dictionary or list with the specified parameters.

    Validators can use the 'param_type' class variable to indicate which type of parameters
    they expect. For collection types like dict or list, it is also possible to specify
    the 'inner_types' attribute to specify requirements on the containing types.
    To understand this feature you probably want to look at the ContainsValidator in
    this file, which uses the 'inner_types' variable.

    Custom Validators should at least overwrite the 'run' method to perform their
    validation operation. Other methods can be overwritten or added on demand. The run
    method should just return if the validation was successful. A validation error
    should be indicated by raising a ValidationException. Errors that are caused by
    invalid configuration within the .yml file should cause a ValidatorError, but
    it is recommended to only raise this error during initialization and not within
    the 'run' method (check the RegexValidator for an example).
    '''
    param_type = None
    inner_types = None

    def __init__(self, path: Path, name: str, param: Any, variables: dict[str, Any]) -> None:
        '''
        Initializes the Validator.

        Parameters:
            path        Path to the .yml file that contains the validator
            name        Name of the Validator (as specified in the .yml file)
            param       Parameters of the Validator (can be str, dict, int, bool, ...)
            variables   Variables contained in this dict are replaced in all str parameters
        '''
        self.path = path
        self.name = name
        self.param = param
        self.variables = variables
        self.check_param_type()

        self.command = None

    def check_param_type(self) -> None:
        '''
        Checks whether the specified parameter type matches the expected one. If this is not
        the case, a ValidatorError is raised.

        Parameters:
            None

        Returns:
            None
        '''
        self.param = tricot.utils.apply_variables(self.param, self.variables)

        if self.param_type is None:
            return

        if type(self.param) is not self.param_type:
            message = f"Validator '{self.name}' requires a parameter type of {str(self.param_type)}"
            raise ValidatorError(self.path, message)

        self.check_inner_types()
        self.check_streams()

    def check_inner_types(self) -> None:
        '''
        Validators that require a dictionary or list can specify the static class variable 'inner_types'
        to specify which inner types are expected within the collection types. When a list is used,
        'inner_types' is expected to contain a list with allowed types. In case of a dict, 'inner_types'
        is expected to contain a dict, containing the possible parameters as keys and a dict of
        {'required': bool, 'type': type} as value.

        The check is currently only implemented on the first layer (non recursive).

        Parameters:
            None

        Returns:
            None
        '''
        if type(self.param) is dict and type(self.inner_types) is dict:

            for key, value in self.inner_types.items():

                param = self.param.get(key)
                if value['required'] and param is None and not self.check_alternative(key):

                    message = f"Validator '{self.name}' requires key with name '{key}' and type {value['type']}."
                    raise ValidatorError(self.path, message)

                if param is not None and type(param) is not value['type']:

                    message = f"Validator '{self.name}' expects type {value['type']} for the '{key}' key."
                    raise ValidatorError(self.path, message)

                self.check_expected()

        elif type(self.param) is list and type(self.inner_types) is list:

            for param in self.param:
                if type(param) not in self.inner_types:
                    message = f"Validator '{self.name}' requires a parameter type of list[{str(self.inner_types)}."
                    raise ValidatorError(self.path, message)

    def check_alternative(self, key: str) -> bool:
        '''
        Checks whether the specified key has alternative values defined. If so, the function
        checks whether one of these is contained within the parameter dictionary.

        Parameters:
            key         Key to look for

        Returns:
            bool        True if one of the alternatives of key is contained
        '''
        alternatives = self.inner_types[key].get('alternatives', [])

        for item in alternatives:
            if item in self.param:
                return True

        return False

    def check_expected(self) -> None:
        '''
        For validators that accept dictionaries, it is possible that typos mess up the test configuration.
        E.g. when using the 'ignore_case' key within the ContainsValidator, a user might misspell it, and
        the check is performed case sensitive instead. This function validates (for Validators using the
        inner_types attribute), that there are no keys that are not contained within the inner_types dict.

        Parameters:
            None

        Returns:
            None
        '''
        if type(self.param) is dict and type(self.inner_types) is dict:

            expected_keys = set(self.inner_types.keys())
            expected_keys.add('stream')

            for key in self.param.keys():

                if key != 'description' and key not in expected_keys:
                    tricot.logging.Logger.print_yellow('Warning:', end=' ')
                    tricot.logging.Logger.print_mixed_blue_plain('Validator', self.name, 'contains unexpected key', end=': ')
                    tricot.logging.Logger.print_yellow_plain(key, end=' ')
                    print(f'({self.path})')

    def from_list(path: Path, input_list: dict, variables: dict[str, Any] = {}) -> list[Validator]:
        '''
        Constructs a new list of Validators from a python list. The list content is usually
        the specified validator configuration within a .yml file. Notice that it is tempting
        to use only a dictionary for validators, as the list seems to be unecessary in most cases.
        However, when you want to use the same validator twice for one test, you need a list.

        Parameters:
            path        Path to the .yml file that contains the validator
            input_list  Python list as read in from a .yml file
            variables   Global variables that should be inherited by the validator

        Returns:
            Validator   Newly constructed Validator objects
        '''
        if type(input_list) is dict:
            raise ValidatorError(path, "Validators need to be specified as a list.")

        validators = list()

        for item in input_list:
            for key, value in item.items():
                val = get_validator(path, key, value, variables)
                validators.append(val)

        return validators

    def resolve_path(self, path: str) -> str:
        '''
        Resolves the specified path relative to the current validator definition (file location of
        the validators .yml file). If the input path is already absolute, it is just returned.

        Parameters:
            path        File system path to resolve

        Returns:
            resolved    Resolved file system path
        '''
        if os.path.isabs(path):
            return path

        return str(self.path.parent.joinpath(path).resolve())

    def check_streams(self) -> None:
        '''
        Used to validate stream names if specified. Is run during parsing of tests to prevent
        errors at runtime. Stream names can be specified for each validator that accepts parameters
        as dictionary values.

        Parameters:
            None

        Returns:
            None
        '''
        if type(self.param) != dict:
            return

        stream = self.param.get('stream')

        if stream is not None:

            if type(stream) != str or stream not in ['stdout', 'stderr', 'both']:
                raise ValidatorError(self.path, "When specified, stream needs to be one of 'stdout', 'stderr' or 'both'.")

    def get_output(self) -> str:
        '''
        Helper function to obtain the actual output of a command execution. Each validator accepts
        the special key 'stream' within of it's parameters. 'stream' is expected to be either 'both',
        'stdout' or 'stderr'. If it was not specified 'both' is taken as the default. The return value
        of this function is the command output that was captured for the correspoding stream.

        Parameters:
            None

        Returns:
            output      Output captured from stdout, stderr or both
        '''
        if type(self.param) != dict:
            return self.command.get_output()

        stream = self.param.get('stream')

        if stream == 'stdout':
            return self.command.stdout

        elif stream == 'stderr':
            return self.command.stderr

        elif stream == 'both' or stream is None:
            return self.command.get_output()

        raise ValidatorError(self.path, f"Encountered unexpected value for the 'stream' key: {stream}")

    def _run(self, command: Command, hotplug_variables: dict[str, Any] = None) -> None:
        '''
        This method is called internally to perform the validation process. It is basically
        a wrapper to the user defined 'run' method that stores the command result in an
        additional class variable.

        Parameters:
            command             Command object associated with the current tests
            hotplug_variables   Variables to apply to the validator at runtime

        Returns:
            None
        '''
        self.command = command
        self.param = tricot.utils.apply_variables(self.param, hotplug_variables)
        self.run()

    def run(self) -> None:
        '''
        Dummy run method. This one needs to be overwritten by the actual validator implementations.

        Parameters:
            None

        Returns:
            None
        '''
        return None


class ContainsValidator(Validator):
    '''
    The ContainsValidator checks whether the command output contains the specified string values.
    The values are expected to be stored as list within the 'values' key. Additionally, the Validator
    accepts the 'ignore_case' key, to specify whether the case should be ignored and the 'invert' key
    to invert the match.

    Example:

        validators:
            - contains:
                ignore_case: True
                values:
                    - match this
                    - and this
                invert:
                    - not match this
                    - and this
    '''
    param_type = dict
    inner_types = {
            'ignore_case': {'required': False, 'type': bool},
            'values': {'required': True, 'type': list, 'alternatives': ['invert']},
            'invert': {'required': True, 'type': list, 'alternatives': ['values']}
    }

    def run(self) -> None:
        '''
        Check whether the command output contains the specified string value.
        '''
        cmd_output = self.get_output()

        invert = self.param.get('invert', [])
        values = self.param.get('values', [])
        ignore_case = self.param.get('ignore_case', False)

        if ignore_case:
            cmd_output = cmd_output.lower()

        for value in values:

            if ignore_case:
                value = value.lower()

            if value not in cmd_output:
                raise ValidationException(f"String '{value}' was not found in command output.")

        for value in invert:

            if ignore_case:
                value = value.lower()

            if value in cmd_output:
                raise ValidationException(f"String '{value}' was found in command output.")


class MatchValidator(Validator):
    '''
    The MatchValidator checks whether there is an exact match of the command output and the
    specified value. The value is expected within the 'value' key. Additionally, the validator
    accepts the 'ignore_case' key, to specify whether the case should be ignored.

    Example:

        validators:
            - match:
                ignore_case: True
                value: Match this!
    '''
    param_type = dict
    inner_types = {
            'ignore_case': {'required': False, 'type': bool},
            'value': {'required': True, 'type': str}
    }

    def run(self) -> None:
        '''
        Check whether command output matches the specified value.
        '''
        value = self.param['value'].rstrip('\n')
        cmd_output = self.get_output().rstrip('\n')

        if self.param.get('ignore_case') is True:
            if value.lower() != cmd_output.lower():
                raise ValidationException(f"String '{value}' does not match command output.")

        else:
            if value != cmd_output:
                raise ValidationException(f"String '{value}' does not match command output.")


class RegexValidator(Validator):
    '''
    The RegexValidator checks whether the command output matches the specified regex.

    Example:

        validators:
            - regex:
                match:
                    - ^match this$
                    - ^and this.+
                invert:
                    - but not this.*
                    - ^or this.*
                multiline: true
                ignore_case: true

    '''
    param_type = dict
    inner_types = {
            'ascii': {'required': False, 'type': bool},
            'dotall': {'required': False, 'type': bool},
            'ignore_case': {'required': False, 'type': bool},
            'multiline': {'required': False, 'type': bool},
            'match': {'required': True, 'type': list, 'alternatives': ['invert']},
            'invert': {'required': True, 'type': list, 'alternatives': ['match']}
    }

    def __init__(self, *args, **kwargs) -> None:
        '''
        Regex initializer. This is used to throw ValidationErrors in case of invalid
        regex specifications already on creation time.
        '''
        super().__init__(*args, **kwargs)

        flags = 0
        self.match = list()
        self.invert = list()

        if self.param.get('ascii'):
            flags = flags | re.ASCII
        if self.param.get('dotall'):
            flags = flags | re.DOTALL
        if self.param.get('ignore_case'):
            flags = flags | re.IGNORECASE
        if self.param.get('multiline'):
            flags = flags | re.MULTILINE

        last = ''

        try:

            for expr in self.param.get('match', []):
                last = expr
                self.match.append(re.compile(expr, flags))

            for expr in self.param.get('invert', []):
                last = expr
                self.invert.append(re.compile(expr, flags))

        except Exception:
            raise ValidatorError(self.path, f"Specified regex '{last}' is invalid!")

    def run(self) -> None:
        '''
        Check whether command output contains the specified regex.
        '''
        cmd_output = self.get_output()

        for regex in self.match:
            if not regex.search(cmd_output):
                raise ValidationException(f"Regex '{regex.pattern}' was not found in command output.")

        for regex in self.invert:
            if regex.search(cmd_output):
                raise ValidationException(f"Regex '{regex.pattern}' was found in command output.")


class StatusCodeValidator(Validator):
    '''
    The StatusCodeValidator checks whether the exit code of the command matches the specified value.

    Example:

        validators:
            - status: 0
    '''
    param_type = int

    def run(self) -> None:
        '''
        Check whether the exit code of the command matches the specified value.
        '''
        status = self.command.status

        if self.param != status:
            raise ValidationException(f"Obtained status code '{status}' does not match the expected code '{self.param}'.")


class ErrorValidator(Validator):
    '''
    The ErrorValidator just checks the status code of the command for an error (status_code != 0). If the
    Validator was used with False as argument, it checks the other way around.

    Example:

        validators:
            - error: False
    '''
    param_type = bool

    def run(self) -> None:
        '''
        Check that status code of command is not equal 0.
        '''
        status = self.command.status

        if status == 0 and self.param is True:
            raise ValidationException("Obtained no error, despite error was expected.")

        if status != 0 and self.param is False:
            raise ValidationException("Obtained error, despite no error was expected.")


class FileExistsValidator(Validator):
    '''
    The FileExistsValidator takes a list of filenames and checks whether they exist on the file
    system. This can be useful, when your testing command is expected to create files.

    Example:

        validators:
            - file_exists:
                cleanup: True
                files:
                    - /tmp/test1
                    - /tmp/test2
                    ...
    '''
    param_type = dict
    inner_types = {
            'cleanup': {'required': False, 'type': bool},
            'files': {'required': True, 'type': list, 'alternatives': ['invert']},
            'invert': {'required': True, 'type': list, 'alternatives': ['files']}
    }

    def run(self) -> None:
        '''
        Check the existence of files and optionally clean them up.
        '''
        not_found = list()

        files = self.param.get('files', [])
        invert = self.param.get('invert', [])

        for file_name in files:

            file_name = self.resolve_path(file_name)
            if os.path.isfile(file_name):

                if self.param.get('cleanup') is True:
                    os.remove(file_name)

                continue

            not_found.append(file_name)

        if not_found:
            file_names = ', '.join(not_found)
            raise ValidationException(f"File '{file_names}' does not exist.")

        for file_name in invert:
            file_name = self.resolve_path(file_name)
            if os.path.isfile(file_name):
                raise ValidationException(f"File '{file_name}' does exist.")


class DirectoryExistsValidator(Validator):
    '''
    The DirectoryExistsValidator takes a list of directory names and checks whether they exist
    on the file system. This can be useful, when your testing command is expected to create
    directories.

    Example:

        validators:
            - dir_exists:
                cleanup: True
                force: False
                dirs:
                    - /tmp/test1
                    - /tmp/test2
                    ...
    '''
    param_type = dict
    inner_types = {
            'cleanup': {'required': False, 'type': bool},
            'force': {'required': False, 'type': bool},
            'dirs': {'required': True, 'type': list, 'alternatives': ['invert']},
            'invert': {'required': True, 'type': list, 'alternatives': ['dirs']}
    }

    def run(self) -> None:
        '''
        Check the existence of directories and optionally clean them up.
        '''
        not_found = list()

        dirs = self.param.get('dirs', [])
        invert = self.param.get('invert', [])

        for dir_name in dirs:

            dir_name = self.resolve_path(dir_name)
            if os.path.isdir(dir_name):

                if self.param.get('cleanup', False):

                    try:
                        os.rmdir(dir_name)

                    except OSError as e:

                        if 'Directory not empty' in str(e) and self.param.get('force', False):
                            shutil.rmtree(dir_name)

                continue

            not_found.append(dir_name)

        if not_found:
            dir_names = ', '.join(not_found)
            raise ValidationException(f"Directory '{dir_names}' does not exist.")

        for dir_name in invert:
            dir_name = self.resolve_path(dir_name)
            if os.path.isdir(dir_name):
                raise ValidationException(f"Directory '{dir_name}' does exist.")


class FileContainsValidator(Validator):
    '''
    The FileContainsValidator takes a list of filenames and strings that represent their
    expected content. It when validates whether the files contain the corresponding content.

    Example:

        validators:
            - file_contains:
                - file: /etc/passwd
                  contains:
                    - root
                    - bin
                  invert:
                    - unexpected
                - file: /etc/hosts
                  ignore_case: True
                  contains:
                    - localhost
                    - 127.0.0.1
    '''
    param_type = list
    inner_types = [dict]

    def run(self) -> None:
        '''
        Check the specified files for their expected content.
        '''
        for check in self.param:

            invert = check.get('invert', [])
            contains = check.get('contains', [])
            ignore_case = check.get('ignore_case', False)

            file_name = self.resolve_path(check['file'])
            if not os.path.isfile(file_name):
                raise ValidationException(f"Specified file '{file_name}' does not exist.")

            with open(file_name) as f:
                content = f.read()

                if ignore_case:
                    content = content.lower()
                    contains = list(map(lambda x: x.lower(), contains))
                    invert = list(map(lambda x: x.lower(), invert))

                for item in contains:
                    if item not in content:
                        raise ValidationException(f"String '{item}' was not found in '{file_name}'.")

                for item in invert:
                    if item in content:
                        raise ValidationException(f"String '{item}' was found in '{file_name}'.")


class RuntimeValidator(Validator):
    '''
    The RuntimeValidator checks whether the overall runtime of the command was lower or greater
    than the user specified value. It also supports equal, but this should be useles :D

    Example:

        validators:
            - runtime:
                lt: 10
                gt: 5
    '''
    param_type = dict
    inner_types = {
            'lt': {'required': True, 'type': int, 'alternatives': ['gt', 'eq']},
            'gt': {'required': True, 'type': int, 'alternatives': ['lt', 'eq']},
            'eq': {'required': True, 'type': int, 'alternatives': ['lt', 'gt']},
    }

    def run(self) -> None:
        '''
        Check that the runtime fulfills the user specified condition.
        '''
        for operation in self.param.keys():

            if operation == 'lt':
                if self.command.runtime >= self.param['lt']:
                    expected = f"(expected: runtime < {self.param['lt']}s)"
                    raise ValidationException(f"Command execution took {self.command.runtime}s {expected}")

            if operation == 'gt':
                if self.command.runtime <= self.param['gt']:
                    expected = f"(expected: runtime > {self.param['gt']}s)"
                    raise ValidationException(f"Command execution took {self.command.runtime}s {expected}")

            if operation == 'eq':
                if self.command.runtime != self.param['eq']:
                    expected = f"(expected: runtime == {self.param['eq']}s)"
                    raise ValidationException(f"Command execution took {self.command.runtime}s {expected}")


class CountValidator(Validator):
    '''
    Takes several strings as argument and an expected count on how often the string should be
    encountered within the command output. Theoretically, we could use the syntax 'match: count'
    to specify that 'match' should be appear 'count' times, as the YAML spec is not that strict
    when it comes to allowed characters within of keys. However, to prevent problems with the python
    parser, two separate lists are used.

    Example:

        validators:
            - count:
                ignore_case: True
                values:
                    - match one
                    - match two
                counts:
                    - 3
                    - 4
    '''
    param_type = dict
    inner_types = {
            'counts': {'required': True, 'type': list},
            'ignore_case': {'required': False, 'type': bool},
            'values': {'required': True, 'type': list}
    }

    def __init__(self, *args, **kwargs) -> None:
        '''
        Count initializer. This makes sure that the values list has the same length
        as the count list.
        '''
        super().__init__(*args, **kwargs)

        if len(self.param['counts']) != len(self.param['values']):
            raise ValidatorError(self.path, "The 'counts' and 'values' lists need to have the same size.")

        for value in self.param['counts']:
            if type(value) != int:
                raise ValidatorError(self.path, "The 'counts' list can only contain numeric items.")

        for value in self.param['values']:
            if type(value) != str:
                raise ValidatorError(self.path, "The 'values' list can only contain string items.")

    def run(self) -> None:
        '''
        Check that the counts are matching.
        '''
        output = self.get_output()
        if self.param.get('ignore_case'):
            output = output.lower()

        for value, count in zip(self.param['values'], self.param['counts']):

            if self.param.get('ignore_case'):
                value = value.lower()

            n = output.count(value)
            if count != n:
                raise ValidationException(f"String '{value}' was found {n} times, but was expected {count} times.")


class LineCountValidator(Validator):
    '''
    The LineCountValidator checks whether the command output contains the expected number of lines.

    Example:

        validators:
            - line_count:
                count: 5
                ignore_empty: True
    '''
    param_type = dict
    inner_types = {
            'count': {'required': True, 'type': int},
            'ignore_empty': {'required': False, 'type': bool},
            'keep_trailing': {'required': False, 'type': bool},
            'keep_leading': {'required': False, 'type': bool},
    }

    def run(self) -> None:
        '''
        Check whether the exit code of the command matches the specified value.
        '''
        count = self.param['count']
        output = self.get_output()
        lines = output.split('\n')

        if not self.param.get('keep_leading'):

            for item in lines[:]:

                if item != '':
                    break

                lines.pop(0)

        if not self.param.get('keep_trailing'):

            for item in lines[::-1]:

                if item != '':
                    break

                lines.pop(-1)

        if self.param.get('ignore_empty'):
            lines = list(filter(lambda x: x, lines))

        if len(lines) != count:
            raise ValidationException(f"Command output has '{len(lines)}' line(s), but '{count}' lines were expected.")


register_validator("contains", ContainsValidator)
register_validator("match", MatchValidator)
register_validator("regex", RegexValidator)
register_validator("status", StatusCodeValidator)
register_validator("error", ErrorValidator)
register_validator("file_exists", FileExistsValidator)
register_validator("dir_exists", DirectoryExistsValidator)
register_validator("file_contains", FileContainsValidator)
register_validator("runtime", RuntimeValidator)
register_validator("count", CountValidator)
register_validator("line_count", LineCountValidator)



from __future__ import annotations

import sys
import json
from termcolor import cprint

from tricot.command import Command
from tricot.validation import Validator, ValidationException, ValidatorError


class Logger:
    '''
    A very primitive Logger class to unify indentation, colors and prefixes.
    In future, this class could also implement additional logging strategies
    like e.g. printing to a file and so on. For now, we keep it simple :)

    The class has two static class variables. The 'indent' variable contains
    the current indentation level of print statements that are executed by
    the Logger. The 'verbosity' variable determines how verbose failed
    Validators should behave.

        Value: 0 -> Just print 'failed' without details
        Value: 1 -> Print 'failed' and show the Validator message
        Value: 2 -> Print 'failed', the validator message + cmd_output
    '''
    tee = None
    indent = 0
    verbosity = 1

    def print(string: str, *args, e: bool = False, end: str = None) -> None:
        '''
        Print with prefix and indentation.
        '''
        print(Logger.get_prefix(e), string, *args, end=end)

    def print_plain(string: str, *args, end: str = None) -> None:
        '''
        Print with prefix.
        '''
        print(string, *args, end=end)

    def print_plain_green(string: str, end: str = None) -> None:
        '''
        Print without indent or prefix, text in green.
        '''
        cprint(string, color='green', end=end)

    def print_plain_red(string: str, end: str = None) -> None:
        '''
        Print without indent or prefix, text in red.
        '''
        cprint(string, color='red', end=end)

    def print_yellow(string: str, e: bool = False, end: str = None) -> None:
        '''
        Print with prefix and indent, text in yellow.
        '''
        print(Logger.get_prefix(e), end=' ')
        cprint(string, color='yellow', end=end)

    def print_yellow_plain(string: str, e: bool = False, end: str = None) -> None:
        '''
        Print in yellow.
        '''
        cprint(string, color='yellow', end=end)

    def print_blue(string: str, e: bool = False, end: str = None, **kwargs) -> None:
        '''
        Print with prefix and indent, text in blue.
        '''
        print(Logger.get_prefix(e), end=' ', **kwargs)
        cprint(string, color='blue', end=end, **kwargs)

    def print_mixed_yellow(str1: str, str2: str, *args, e: bool = False,  end: str = None) -> None:
        '''
        Print with prefix and indent, first arg in normal text color, second arg
        in yellow and the rest of the args in normal text color again.
        '''
        print(Logger.get_prefix(e), str1, end=' ')

        if(args):
            cprint(str2, color='yellow', end=' ')
            print(*args, end=end)

        else:
            cprint(str2, color='yellow', end=end)

    def print_mixed_red(str1: str, str2: str, *args, e: bool = False,  end: str = None) -> None:
        '''
        Print with prefix and indent, first arg in normal text color, second arg
        in red and the rest of the args in normal text color again.
        '''
        print(Logger.get_prefix(e), str1, end=' ')

        if(args):
            cprint(str2, color='red', end=' ')
            print(*args, end=end)

        else:
            cprint(str2, color='red', end=end)

    def print_mixed_blue(str1: str, str2: str, *args, e: bool = False, end: str = None) -> None:
        '''
        Print with prefix and indent, first arg in normal text color, second arg
        in blue and the rest of the args in normal text color again.
        '''
        print(Logger.get_prefix(e), str1, end=' ')

        if(args):
            cprint(str2, color='blue', end=' ')
            print(*args, end=end)

        else:
            cprint(str2, color='blue', end=end)

    def print_mixed_blue_plain(str1: str, str2: str, *args, e: bool = False, end: str = None) -> None:
        '''
        Print first arg in normal text color, second arg
        in blue and the rest of the args in normal text color again.
        '''
        print(str1, end=' ')

        if(args):
            cprint(str2, color='blue', end=' ')
            print(*args, end=end)

        else:
            cprint(str2, color='blue', end=end)

    def print_mixed_red_plain(str1: str, str2: str, *args, e: bool = False, end: str = None) -> None:
        '''
        Print first arg in normal text color, second arg
        in red and the rest of the args in normal text color again.
        '''
        print(str1, end=' ')

        if(args):
            cprint(str2, color='red', end=' ')
            print(*args, end=end)

        else:
            cprint(str2, color='red', end=end)

    def increase_indent() -> None:
        '''
        Increases the indentation level by one.
        '''
        Logger.indent += 1

    def decrease_indent() -> None:
        '''
        Decreases the indentation level by one.
        '''
        Logger.indent -= 1

        if Logger.indent < 0:
            Logger.indent = 0

    def reset_indent() -> None:
        '''
        Resets the indent level to 0.
        '''
        Logger.indent = 0

    def set_verbosity(level: int) -> None:
        '''
        Sets the verbosity level to the specified value
        '''
        Logger.verbosity = level

    def get_prefix(is_error: bool) -> None:
        '''
        Returns the globally used prefix. '[+]' on success and '[-]' on error.
        '''
        if is_error:
            return '[-]' + Logger.indent * 4 * ' '

        return '[+]' + Logger.indent * 4 * ' '

    def print_with_indent(string: str, e: bool = False) -> None:
        '''
        Takes a string, spilts it on newlines and prints each single line
        with the current prefix and indent.
        '''
        content = string.replace('\x0d', '\n')
        lines = content.split('\n')
        for line in lines:
            Logger.print(line, e=e)

    def print_with_indent_blue(string: str, e: bool = False) -> None:
        '''
        Takes a string, spilts it on newlines and prints each single line
        with the current prefix and indent in blue.
        '''
        content = string.replace('\x0d', '\n')
        lines = content.split('\n')
        for line in lines:
            Logger.print_blue(line, e=e)

    def set_logfile(path: str) -> None:
        '''
        Mirrors all output of tricot to the specified logfile.

        Parameters:
            path        File system path of the logfile

        Returns:
            None
        '''
        Logger.tee = Tee(path)

    def handle_error(e: Exception, val: Validator) -> None:
        '''
        This function is called when a Validator failed to handle the error
        logging. It is basically a wrapper around '_handle_error', which
        performs the real error handling.
        '''
        Logger.increase_indent()
        Logger._handle_error(e, val)
        Logger.decrease_indent()

        if Logger.verbosity != 0:
            Logger.print('', e=True)

    def _handle_error(e: Exception, val: Validator) -> None:
        '''
        This functions handles Exceptions that are raised by validators.
        Depending on the current 'verbosity' level, a different amount of information
        is printed.
        '''
        if Logger.verbosity == 0:
            return

        if type(e) is ValidatorError:
            Logger.print_mixed_red('- Caught', 'ValidatorError', 'during validator instantiation.', e=True)
            Logger.print('  Validator instantiation failed because of the following reason:', e=True)
            Logger.print_blue('  ' + str(e), e=True)
            return

        elif type(e) is ValidationException:
            Logger.print_mixed_yellow('- Caught', 'ValidationException', 'raised by the', end='', e=True)
            Logger.print_mixed_red_plain('', val.name, 'validator.')
            Logger.print_mixed_blue('  Configuration file:', val.path.absolute(), e=True)

            if Logger.verbosity > 1:
                Logger.print('', e=True)

            Logger.print('  Validator run failed because of the following reason:', e=True)
            Logger.print_with_indent_blue('  ' + str(e), e=True)

        else:
            Logger.print_mixed_yellow('- Caught unexpected', type(e).__name__, 'during validation process.', e=True)
            Logger.print_mixed_red('  The', val.name, 'validator raises probably an uncaught exception.', e=True)
            Logger.print_mixed_blue('  Message:', str(e), e=True)

        if Logger.verbosity > 1:
            Logger.print('', e=True)
            Logger.print_yellow('  Command:', e=True, end=' ')
            Logger.print_plain(val.command.command)
            Logger.print_yellow('  Command exit code:', end=' ', e=True)
            cprint(val.command.status, color='blue')

            Logger.print_yellow('  Command stdout:', e=True)
            if val.command.stdout:
                Logger.increase_indent()
                Logger.print_with_indent(val.command.stdout, e=True)
                Logger.decrease_indent()

            Logger.print_yellow('  Command stderr:', e=True)
            if val.command.stderr:
                Logger.increase_indent()
                Logger.print_with_indent(val.command.stderr, e=True)
                Logger.decrease_indent()

            Logger.print_yellow('  Validator parameters:', e=True)
            Logger.increase_indent()
            Logger.print_with_indent(json.dumps(val.param, indent=4).replace(r'\n', '\n').replace(r'\t', '\t'), e=True)
            Logger.decrease_indent()

    def handle_success(cmd: Command, val: list[Validator]) -> None:
        '''
        This function is called when all Validators have been passed.
        It is basically a wrapper around '_handle_success'.
        '''
        if Logger.verbosity != 3:
            return

        Logger.increase_indent()
        Logger._handle_success(cmd, val)
        Logger.decrease_indent()
        Logger.print('')

    def _handle_success(cmd: Command, val: list[Validator]) -> None:
        '''
        This functions handles Exceptions that are raised by validators.
        Depending on the current 'verbosity' level, a different amount of information
        is printed.
        '''
        Logger.print_mixed_yellow('-', 'Debug output:')
        Logger.increase_indent()
        Logger.print_yellow('Command:', end=' ')
        Logger.print_plain(cmd.command)
        Logger.print_yellow('Command exit code:', end=' ')
        cprint(cmd.status, color='blue')

        Logger.print_yellow('Command stdout:')
        if cmd.stdout:
            Logger.increase_indent()
            Logger.print_with_indent(cmd.stdout)
            Logger.decrease_indent()

        Logger.print_yellow('Command stderr:')
        if cmd.stderr:
            Logger.increase_indent()
            Logger.print_with_indent(cmd.stderr)
            Logger.decrease_indent()

        Logger.print_yellow('Validators:')
        Logger.increase_indent()
        for validator in val:
            Logger.print_yellow('Validator name:', end=' ')
            cprint(validator.name, color='blue')
            Logger.print_yellow('Validator parameters:')
            Logger.increase_indent()
            Logger.print_with_indent(json.dumps(validator.param, indent=4).replace(r'\n', '\n').replace(r'\t', '\t'))
            Logger.decrease_indent()

        Logger.decrease_indent()
        Logger.decrease_indent()


class Tee(object):
    '''
    Helper class to log stdout to a file. Copied from:
    https://stackoverflow.com/questions/616645/how-to-duplicate-sys-stdout-to-a-log-file
    '''
    def __init__(self, name):
        self.file = open(name, 'w')
        self.stdout = sys.stdout
        sys.stdout = self

    def __del__(self):
        sys.stdout = self.stdout
        self.file.close()

    def write(self, data):
        self.file.write(data)
        self.stdout.write(data)

    def flush(self):
        self.file.flush()



from __future__ import annotations

import copy
import yaml
import glob
from pathlib import Path
from typing import Any, Union

import tricot.utils
from tricot.docker import TricotContainer
from tricot.logging import Logger
from tricot.validation import Validator, ValidationException
from tricot.plugin import Plugin
from tricot.command import Command
from tricot.condition import Condition


class TricotException(Exception):
    '''
    Custom exception class for general purpose and tricot related exceptions.
    '''


class TesterKeyError(Exception):
    '''
    TesterKeyErrors are raised when the .yml test definition file misses
    required keys for constructing a Tester object.
    '''

    def __init__(self, key: str, path: Path, optional: str = None, section: str = None) -> None:
        '''
        As these Exception is only used internally, we give it a static content
        that only takes the missing key as argument.
        '''
        if optional is not None:
            self.message = f"Test configuration requires either the '{key}' or the '{optional}' key"
            self.message += "within the 'global' section."

        elif section is not None:
            self.message = f"Test configuration misses required key {key} in the '{section}' section."

        else:
            self.message = f"Test configuration misses required key {key} in the 'global' section."

        self.path = str(path.resolve())
        super().__init__(self.message)


class TestKeyError(Exception):
    '''
    TestKeyErrors are raised when the .yml test definition file misses
    required keys for constructing a Test object.
    '''

    def __init__(self, key: str, path: Path, message: str = None) -> None:
        '''
        As these Exception is only used internally, we give it a static content
        that only takes the missing key as argument. Optionally, we allow to add
        a custom message if required.
        '''
        if message:
            self.message = message

        else:
            self.message = f"Test configuration misses required key {key} within the "

        self.path = str(path.resolve())
        self.number = None

        super().__init__(self.message)

    def add_number(self, ctr: int) -> None:
        '''
        The default content of the exception indicates a missing key within a test configuration.
        This function can be used to add the position of the corresponding test within the .yml
        file.
        '''
        self.number = ctr

    def __str__(self) -> str:
        '''
        Depending whether the execption contains the test number (self.number), the exception
        is formatted a little bit differently.
        '''
        message = self.message

        if self.number:
            ordinal = tricot.utils.make_ordinal(self.number)
            message += f'{ordinal} test in the '

        return message + 'tests section.'


class Test:
    '''
    The Test class represents a single command test specified within a .yml file.
    Test objects contain all required information like the command definition, the
    specified variables and the required validators. During a test, Test objects are
    evaluated by executing their 'run' method.
    '''
    expected_keys = ['title', 'description', 'command', 'arguments', 'validators', 'timeout', 'env', 'conditions']

    def __init__(self, path: Path, title: str, error_mode: str, variables: dict[str, Any], command: list,
                 timeout: int, validators: list[Validator], env: dict, conditions: dict, conditionals: set[Condition]) -> None:
        '''
        Initializer for a Test object.

        Parameters:
            path            Path object to the file that contains the tests configuration
            title           The title of the test - is displayed during execution.
            error_mode      Decides what to do if a validator fails (break|continue)
            variables       Dictionary of variables that are replaced in the command
            command         The command that is run by this test
            timeout         Timeout in seconds to wait for the specified command
            validators      List of validators to apply on the command output
            env             Environment variables
            conditions      Conditions for running the current test
            conditionals    Conditionals defined by upper testers

        Returns:
            None
        '''
        self.path = path
        self.title = title
        self.error_mode = error_mode
        self.variables = variables
        self.command = Command(command)
        self.timeout = timeout
        self.validators = validators
        self.env = env
        self.conditions = conditions
        self.conditionals = conditionals

    def apply_variables(val: Union(str, list), variables: dict[str, Any], k: str = 'command') -> list:
        '''
        Applies variables to a command or argument list/string. This method needs to be
        static, as it is applied right at the object creation. This is the only way to
        allow command lists to be specified within variables while simoultaneously checking
        types on object creation.

        Parameters:
            val             Value to replace variables in. Usually the command or arguments portion of .yml files
            variables       Variables to apply
            k               Current key value (command|arguments)

        Returns:
            None
        '''
        if val is None:
            return []

        for key, value in variables.items():

            value = tricot.utils.resolve_runtime_variables(variables, key, value)
            value = tricot.utils.resolve_env_variables(variables, key, value)
            key = '${'+str(key)+'}'

            if type(val) is str:
                if val == key and type(value) is list:
                    val = value.copy()

                else:
                    val = val.replace(key, str(value))

            elif type(val) is list:
                for ctr in range(len(val)):
                    if type(val[ctr]) is str:
                        val[ctr] = val[ctr].replace(key, str(value))

                    else:
                        val[ctr] = str(val[ctr])

        if type(val) is not list:
            raise ValueError(f"The '{k}' key needs to be a list within the ")

        return val

    def from_dict(path: Path, input_dict: dict, variables: dict[str, Any] = {},
                  error_mode: str = 'continue', environment: dict = {}, conditionals: set[Condition] = set()) -> Test:
        '''
        Creates a Test object from a dictionary. The dictionary is expected to be the content
        read in of a .yml file and needs all keys that are required for a test (validators,
        title and command).

        Parameters:
            path            Path object to the tests configuration file
            input_dict      Dictionary that defines all required information for a test
            variables       Variables that should be inherited from the global scope
            error_mode      Decides whether to break or continue on failure
            environment     Environment variables
            conditionals    Conditionals for running the test

        Returns:
            Test            Newly generated Test object
        '''
        try:
            j = input_dict
            var = {**variables, **j.get('variables', {})}
            validators = Validator.from_list(path, j['validators'], var)

            e_mode = j.get('error_mode') or error_mode
            env = tricot.utils.merge_environment(j.get('env'), environment, path)
            conditions = j.get('conditions', {})

            Condition.check_format(path, conditions, conditionals)

            command = Test.apply_variables(j['command'], var)
            arguments = Test.apply_variables(j.get('arguments'), var, 'arguments')

            if type(command) is not list:
                raise TestKeyError(None, path, "The 'command' key needs to be a list within the ")

            if arguments:

                if type(arguments) is not list:
                    raise TestKeyError(None, path, "The 'arguments' key needs to be a list within the ")

                command += arguments

            tricot.utils.check_keys(Test.expected_keys, input_dict)
            return Test(path, j['title'], e_mode, var, command, j.get('timeout'), validators, env, conditions, conditionals)

        except KeyError as e:
            raise TestKeyError(str(e), path)

        except ValueError as e:
            raise TestKeyError(None, path, str(e))

    def from_list(path: Path, input_list: list, variables: dict[str, Any] = {},
                  error_mode: str = 'continue', env: dict = {}, conditionals: set[Condition] = set()) -> list[Test]:
        '''
        Within .yml files, Tests are specified in form of a list. This function takes such a list,
        that contains each single test definition as another dictionary (like it is created when
        reading the .yml file) and returns a corresponding list of Test objects.

        Parameters:
            path            Path object to the tests configuration file
            input_list      List of test definitions as read in from a .yml file
            variables       Variables that should be inherited from the global scope
            error_mode      Decides whether to break or continue on failure
            env             Environment variables
            conditionals    Conditionals specified by the upper tester

        Returns
            list[Test]      List of Test objects created from the .yml input
        '''
        tests = list()

        if type(input_list) is not list:
            raise TestKeyError(None, path, 'Test defintions need to be specified as list within the ')

        for ctr in range(len(input_list)):

            try:
                test = Test.from_dict(path, input_list[ctr], variables, error_mode, env, conditionals)
                tests.append(test)

            except TestKeyError as e:
                e.add_number(ctr + 1)
                raise e

        return tests

    def run(self, prefix: str = '-', hotplug_variables: dict[str, Any] = None) -> None:
        '''
        Runs the Test and applies all specified validators to the command output.
        Depending on the current error_mode, the function may raise exceptions and
        let the caller handle them. Furthermore, it generates output regarding the
        success or error of a Test.

        Parameters:
            prefix              Optional prefix to include for each test title
            hotplug_variables   Variables that are applied at runtime.

        Returns:
            None
        '''
        Logger.increase_indent()
        Logger.print_blue(f'{prefix} {self.title}...', end=' ', flush=True)
        success = True

        if not Condition.check_conditions(self.conditions, self.conditionals):
            Logger.print_yellow_plain("skipped.")
            Logger.decrease_indent()
            return

        self.command.run(self.path.parent, self.timeout, hotplug_variables, self.env)

        for validator in self.validators:

            try:
                validator._run(self.command, hotplug_variables)

            except Exception as e:

                if success:
                    Logger.print_plain_red("failed.")

                tricot.constants.LAST_ERROR = tricot.constants.VALIDATION_EXCEPTION
                Logger.handle_error(e, validator)
                Condition.update_conditions(self.conditions, self.conditionals, True)

                if self.error_mode == "break":
                    Logger.decrease_indent()
                    raise ValidationException('')

                else:
                    success = False

        if success:
            Condition.update_conditions(self.conditions, self.conditionals, False)
            Logger.print_plain_green("success.")
            Logger.handle_success(self.command, self.validators)

        hotplug_variables['$prev'] = self.command
        Logger.decrease_indent()


class Tester:
    '''
    The Tester class represents a single .yml file that contains Test definitions. It is basically
    an Object that contains the tester configuration and a reference to each Test object that should
    be executed during a run. A command test with tricot is intended to be started by a Tester object,
    by calling its 'run' method.
    '''
    tester_count = 0

    def __init__(self, path: Path, name: str, title: str, variables: dict[str, Any], tests: list[Test], testers: list[Tester],
                 containers: list[TricotContainer], plugins: list[Plugin], conditions: dict, conditionals: set[Condition],
                 error_mode: str) -> None:
        '''
        Initializes a new Tester object.

        Parameters:
            path            Path object to the Testers configuration file
            name            Name of the tester (used for identification)
            title           Title of the test (used for displaying)
            variables       Dictionary of global variables that should be applied to all Tests and Validators
            tests           List of Test objects that should run during a test
            testers         List of Tester object, that allows to nest tester objects
            containers      List of TricotContainer objects for the test
            plugins         List of Plugin objects for the test
            conditions      Conditions for running the current tester
            conditionals    Conditionals defined by the tester or upper testers
            error_mode      Decides what to do if a plugin fails (break|continue)
        '''
        self.name = name
        self.title = title or name
        self.variables = variables
        self.tests = tests
        self.testers = testers
        self.containers = containers
        self.plugins = plugins
        self.conditions = conditions
        self.conditionals = conditionals
        self.error_mode = error_mode

    def from_dict(input_dict: dict, initial_vars: dict[str, Any] = dict(),
                  path: Path = None, e_mode: str = None, environment: dict = {},
                  conditionals: set[Condition] = set()) -> Tester:
        '''
        Creates a new Tester object from a python dictionary. The dictionary is expected to be
        created by reading a .yml file that contains test defintions. It requires all keys that
        need to be present for creating the Tester and the Test objects. This method walks down
        basically the complete .yml tree and constructs all required objects on its way.

        Parameters:
            input_dict      Python dictionary as read in from a .yml file
            initial_vars    Additional variables. This is used internally when nesting Testers
            path            Path object to the testers configuration file
            e_mode          Error mode that was mayve inherited by the parent tester
            environment     Dictionary of environment variables to use within the test
            conditionals    Conditions inherited from the upper tester

        Returns:
            Tester          Tester object created from the dictionary
        '''
        try:
            g = input_dict
            t = input_dict['tester']

            env = tricot.utils.merge_environment(t.get('env'), environment, path)
            error_mode = t.get('error_mode') or e_mode
            conds = Condition.from_dict(path, t.get('conditionals', {})).union(conditionals)
            run_conds = t.get('conditions', {})

            Condition.check_format(path, run_conds, conds)

            testers = g.get('testers')
            definitions = g.get('tests')

            variables = g.get('variables', dict())
            variables = {**initial_vars, **variables}
            variables['cwd'] = path.parent
            variables = tricot.utils.apply_variables(variables, copy.deepcopy(variables))

            plugins = Plugin.from_list(path, g.get('plugins'), variables)
            containers = TricotContainer.from_list(g.get('containers', list()), path, variables)

            tester_list = list()
            if testers and type(testers) is list:

                for f in testers:

                    if not Path(f).is_absolute():
                        f = path.parent.joinpath(f)

                    for ff in glob.glob(str(f)):
                        tester = Tester.from_file(ff, variables, None, error_mode, env, conds)
                        tester_list.append(tester)

            tests = None
            if definitions and type(definitions) is list:
                tests = Test.from_list(path, definitions, variables, error_mode, env, conds)

            elif not tester_list:
                raise TesterKeyError('tests', path, optional='testers')

            return Tester(path, t['name'], t.get('title'), variables, tests, tester_list, containers, plugins,
                          run_conds, conds, error_mode)

        except KeyError as e:
            raise TesterKeyError(str(e), path)

    def from_file(filename: str, initial_vars: dict[str, Any] = dict(), runtime_vars: dict[str, Any] = None,
                  error_mode: str = None, env: dict = {}, conditionals: set[Condition] = set()) -> Tester:
        '''
        Creates a new Tester object from a .yml file. The .yml file obviously needs to be in the
        expected format and requires all keys that are needed to construct a Tester object.

        Parameters:
            filename        File system path to the corresponding .yml file
            initial_vars    Additional variables. This is used internally when nesting Testers
            runtime_vars    Runtime variables (only specifiable when using tricot as library)
            error_mode      Current error mode setting
            env             Current environment variables
            conditionals    Conditions inherited from the previous tester

        Returns:
            Tester          Tester object created from the file
        '''
        with open(filename, 'r') as f:
            config_dict = yaml.safe_load(f.read())

        if '$env' not in initial_vars:
            tricot.utils.add_environment(initial_vars)

        if runtime_vars is not None and '$runtime' not in initial_vars:
            initial_vars['$runtime'] = runtime_vars

        return Tester.from_dict(config_dict, initial_vars, Path(filename), error_mode, env, conditionals)

    def contains_testers(self, testers: list[str]) -> bool:
        '''
        Checks whether the specified tester name is contained within the current test tree.

        Parameters:
            testers         List of tester names to look for

        Returns:
            bool            True if tester is contained within test tree
        '''
        if len(testers) == 0:
            return True

        for tester in testers:

            if self.name == tester:
                return True

        for tester in self.testers:

            if tester.contains_testers(testers):
                return True

        return False

    def run(self, testers: list[str] = (), numbers: list[str] = (), exclude: list[str] = (),
            hotplug_variables: dict[str, Any] = {}) -> None:
        '''
        Runs the test: Prints the title of the tester and iterates over all contained
        Test objects and calls their 'run' method.

        Parameters:
            tester              Only run testers that match the specified names
            numbers             Only run tests that match the specified numbers
            exclude             Exclude the specified tester names
            hotplug_variables   Dictionary of variables that are applied at runtime.

        Returns:
            None
        '''
        if not self.contains_testers(testers) or self.name in exclude:
            return

        Tester.increase()

        if not Condition.check_conditions(self.conditions, self.conditionals):
            Logger.print_mixed_yellow('Skipping test:', self.title)
            return

        Logger.print_mixed_yellow('Starting test:', self.title)
        hotplug = hotplug_variables.copy()

        Logger.increase_indent()

        for plugin in self.plugins:
            plugin._run(hotplug)

        for container in self.containers:
            container.start_container()
            hotplug = {**hotplug, **container.get_container_variables()}

        if self.tests:
            Logger.print('')
            for ctr in range(len(self.tests)):
                if len(numbers) == 0 or (ctr+1) in numbers:
                    self.tests[ctr].run(f'{ctr+1}.', hotplug)

        self.run_childs(testers, numbers, exclude, hotplug)

        for container in self.containers:
            container.stop_container()

        for plugin in self.plugins:
            plugin._stop()

        Logger.decrease_indent()

    def run_childs(self, testers: list[str] = (), numbers: list[str] = (), exclude: list[str] = (),
                   hotplug_variables: dict[str, Any] = {}) -> None:
        '''
        Runs the child testers of the current tester.

        Parameters:
            tester              Only run testers that match the specified names
            numbers             Only run tests that match the specified numbers
            exclude             Exclude the specified tester names
            hotplug_variables   Dictionary of variables that are applied at runtime.

        Returns:
            None
        '''
        for tester in self.testers:
            try:

                if self.name in testers:
                    tester.run(numbers=numbers, exluce=exclude, hotplug_variables=hotplug_variables)

                else:
                    tester.run(testers, numbers, exclude, hotplug_variables)

            except tricot.PluginException as e:

                tricot.Logger.print("")

                if self.error_mode == 'break':
                    raise e

                else:
                    tricot.Logger.print_mixed_yellow('Caught', 'PluginException', 'from', end='', e=True)
                    tricot.Logger.print_mixed_blue_plain('', e.name, 'plugin in', end=' ')
                    tricot.Logger.print_yellow_plain(e.path.absolute())
                    tricot.Logger.print_mixed_blue('Original exception:', f'{type(e.original).__name__} - {e.original}')
                    tricot.Logger.print_blue('Tester is skipped.', e=True)

                    tricot.constants.LAST_ERROR = tricot.constants.PLUGIN_EXCEPTION

    def increase() -> None:
        '''
        Helper function that increases the global tester count and prints an offset
        if the current tester is not the first one.
        '''
        if Tester.tester_count != 0:
            Logger.print('')

        Tester.tester_count += 1



import tricot


tester_template = """tester:
  name: <NAME>
  title: <TITLE>
  description: |-
    <DESCRIPION>


variables:
  <VAR>: <VALUE>


containers:
  - name: <NAME>
    image: <IMAGE>
    volumes:
      - <LOCAL>:<CONTAINER>
    env:
        <KEY>: <VALUE>
    aliases:
      DOCKER-<NAME>-IP: DOCKER-IP
      DOCKER-<NAME>-GATEWAY: DOCKER-GW


plugins:
  - <PLUGIN>


tests:
  - title: <TITLE>
    description: |-
        <DESCRIPTION>

    command:
      - <CMD>
      - <ARGG1>

    validators:
      - <VALIDATOR>


testers:
  - <TESTER>
"""


plugin_template = """import tricot

class MyPlugin(tricot.Plugin):
    '''
    Plugin description.

    Example:

        plugins:
            - my_plugin:
                key: value
    '''
    param_type = dict
    inner_types = {
                    'key': {'required': True, 'type': str},
                  }

    def run(self) -> None:
        '''
        Executed on startup.
        '''
        pass

    def stop(self) -> None:
        '''
        Executed when stopping.
        '''
        pass


tricot.register_plugin('my_plugin', MyPlugin)
"""


validator_template = """import tricot

class MyValidator(tricot.Validator):
    '''
    Validator description.

    Example:

        validators:
            - my_validator:
                key: value
    '''
    param_type = dict
    inner_types = {
            'key': {'required': True, 'type': str},
    }

    def run(self) -> None:
        '''
        Run during validation.
        '''
        if self.command.status != 0:
            raise tricot.ValidationException(f'Failure Reason')

        if 'key' not in self.get_output():
            raise tricot.ValidationException(f'Failure Reason')


tricot.register_validator('my_validator', MyValidator)
"""


def replace_placeholders(template: str) -> str:
    '''
    Replaces the palceholders within a template. Accepted placeholders
    are currently the plugin and validator names.

    Parameters:
        template    YAML template to replace parameters in

    Returns:
        YAML template with placeholders replaced
    '''
    val_list = tricot.get_validator_list()
    val_replace = '\n      - '.join(val_list)

    replaced = template.replace('<VALIDATOR>', val_replace)

    plug_list = tricot.get_plugin_list()
    plug_replace = '\n  - '.join(plug_list)

    replaced = replaced.replace('<PLUGIN>', plug_replace)
    return replaced


def write_template(filename: str, mode: str) -> None:
    '''
    Write the template YAML to the specified file system path.

    Parameters:
        filename    file system location to write to
        mode        Decdides whether the tester, plugin or validator template is written

    Returns:
        None
    '''
    if mode.lower() == 'tester':
        prepared_template = replace_placeholders(tester_template)

    elif mode.lower() == 'plugin':
        prepared_template = plugin_template

    elif mode.lower() == 'validator':
        prepared_template = validator_template

    with open(filename, 'w') as f:
        f.write(prepared_template)



from __future__ import annotations

from pathlib import Path


class ConditionFormatException(Exception):
    '''
    Raised if the condition format in the .yml file was not correct.
    '''
    def __init__(self, message: str, path: Path) -> None:
        '''
        Custom exception class that stores the original exception within a variable.
        '''
        self.path = path
        super().__init__(message)


class Condition:
    '''
    Conditions can be configured within of testers and can be enabled
    or disabled within of tests. Tests can be configured to only run
    if certain conditions are matched.
    '''

    def __init__(self, name: str, state: bool = False) -> None:
        '''
        Initializes the Condition object with a name and an initial
        state.

        Parameters:
            name            Name of the condition
            state           Initial state of the condition

        Returns:
            None
        '''
        self.name = name
        self.state = state

    def enable(self) -> None:
        '''
        Sets the current condition state to True.

        Parameters:
            None

        Returns:
            None
        '''
        self.state = True

    def disable(self) -> None:
        '''
        Sets the current condition state to False.

        Parameters:
            None

        Returns:
            None
        '''
        self.state = False

    def is_enabled(self) -> bool:
        '''
        Returns the current state of the Condition.

        Parameters:
            None

        Returns:
            None
        '''
        return self.state

    def __eq__(self, other) -> bool:
        '''
        Two Conditions are equal, if they have the same name.

        Parameters:
            other       Condition object to compare with

        Returns:
            bool
        '''
        if type(other) is not Condition:
            return False

        return self.name == other.name

    def __hash__(self) -> int:
        '''
        According to __eq__, hashes are also computed on the condition name.

        Parameters:
            None

        Returns:
            hash
        '''
        return hash(self.name)

    def contains_str(condition: str, conditionals: set[Condition]) -> bool:
        '''
        Checks whether {condition} name is contained within the specified
        set of conditions.

        Parameters:
            condition       Condition name to look for
            conditionals    Set of conditions to look in

        Returns:
            result          True if condition was found in conditionals
        '''
        for cond in conditionals:

            if cond.name == condition:
                return True

        return False

    def update_by_str(condition: str, conditionals: set[Condition], value: bool) -> None:
        '''
        Checks the conditionals set for the specified condition string and updates
        the state of the corresponding condition to the specified value.

        Parameters:
            condition       Condition name to look for
            conditionals    Set of conditions to look in
            value           New state value

        Returns:
            None
        '''
        for cond in conditionals:

            if cond.name == condition:
                cond.state = value
                return

    def from_dict(path: Path, input_dict: dict) -> set[Condition]:
        '''
        Creates a set of Conditions from an input dictionary.

        Parameters:
            path            Path to the current .yml file
            input_dict      Input dictionary read from the .yml file

        Returns:
            set             Set of specified Conditions
        '''
        if type(input_dict) is not dict:
            raise ConditionFormatException('Conditions need to be specified as string -> bool pairs', path)

        conditions = set()

        for key, value in input_dict.items():

            if type(key) != str or type(value) != bool:
                raise ConditionFormatException('Conditions need to be specified as string -> bool pairs', path)

            cond = Condition(key, value)
            conditions.add(cond)

        return conditions

    def check_format(path: Path, conditions: dict, conditionals: set[Condition]) -> None:
        '''
        Check that condition section has the correct format within a .yml file. If the format
        is malformed, an exception is raised.

        Parameters:
            path            Path to the .yml file
            conditions      Condition section specified as dictionary
            conditionals    Set of conditions specified by upper testers

        Returns:
            None
        '''
        if conditions is None:
            return True

        cond_all = conditions.get('all', [])
        cond_one_of = conditions.get('one_of', [])
        cond_none_of = conditions.get('none_of', [])

        cond_error = conditions.get('on_error', {})
        cond_success = conditions.get('on_success', {})

        if not all(isinstance(x, list) for x in [cond_all, cond_one_of, cond_none_of]):
            raise ConditionFormatException("The keys 'all', 'one_of' and 'none_of' need to be lists", path)

        if not all(isinstance(x, dict) for x in [cond_error, cond_success]):
            raise ConditionFormatException("The keys 'on_error' and 'on_success' need to be dicts", path)

        for item in set(cond_all + cond_one_of + cond_none_of):

            if not Condition.contains_str(item, conditionals):
                raise ConditionFormatException(f"Condition '{item}' was used but never declared within a tester.", path)

        for current_dict in [cond_error, cond_success]:

            for key, value in current_dict.items():

                if not Condition.contains_str(key, conditionals):
                    raise ConditionFormatException(f"Condition '{key}' was used but never declared within a tester.", path)

                elif type(value) is not bool:
                    raise ConditionFormatException(f"Condition '{key}' is set to a non boolean value.", path)

    def check_conditions(conditions: dict, conditionals: set[Condition]) -> bool:
        '''
        Checks whether the specified conditions allow a run.

        Parameters:
            conditions      Condition section specified as dictionary
            conditionals    Set of conditions specified by upper testers

        Returns:
            None
        '''
        cond_all = conditions.get('all', [])
        cond_one_of = conditions.get('one_of', [])
        cond_none_of = conditions.get('none_of', [])

        conds = set(filter(lambda x: x.is_enabled(), conditionals))

        if cond_one_of:
            c_set = set(map(lambda x: x.name, conds))
            if bool(set(cond_one_of).isdisjoint(c_set)):
                return False

        for cond in cond_all:
            if not Condition.contains_str(cond, conds):
                return False

        for cond in cond_none_of:
            if Condition.contains_str(cond, conds):
                return False

        return True

    def update_conditions(conditions: dict, conditionals: set[Condition], failure: bool) -> None:
        '''
        Updates the specified conditionals according to the specified conditions.

        Parameters:
            conditions      Condition section specified as dictionary
            conditionals    Set of conditions specified by upper testers
            failure         Indicates whether the corresponding test failed or succeeded

        Returns:
            None
        '''
        cond_error = conditions.get('on_error', {})
        cond_success = conditions.get('on_success', {})

        if failure:
            current_dict = cond_error
        else:
            current_dict = cond_success

        for key, value in current_dict.items():
            Condition.update_by_str(key, conditionals, value)



from __future__ import annotations

import os
import tricot
from typing import Any
from pathlib import Path


class TricotRuntimeVariableError(Exception):
    '''
    A TricotRuntimeVariableError is raised, when a runtime variable was specified within
    the .yml test configuration, but was not defined at runtime.
    '''
    pass


class TricotEnvVariableError(Exception):
    '''
    A TricotEnvVariableError is raised, when an environment variable was specified within
    the .yml test configuration, but was not found at runtime.
    '''
    pass


def resolve_runtime_variables(variables: dict[str, Any], key: str, value: Any) -> Any:
    '''
    This function is responsible for resolving runtime variables.
    When a variable contains the value '$runtime' it checks the
    '$runtime' key within the variables dict for the corresponding
    dictionary key. If it is found, the corresponding value is
    returned. If not, a TricotRuntimeVariableError is raised.

    When the value of a variables is not '$runtime', it is directly
    returned.

    Parameters:
        variables       Variables dictionary created during the Tester initialization
        key             Key of the current variable in question
        value           Value of the current variable in question

    Returns:
        value       Value with possible runtime variables applied
    '''
    if value != '$runtime':
        return value

    try:
        return (variables['$runtime'])[key]

    except KeyError:
        raise TricotRuntimeVariableError(f"Unable to find runtime variable '{key}'.")


def resolve_env_variables(variables: dict[str, Any], key: str, value: Any) -> Any:
    '''
    This function is responsible for resolving environment variables.
    When a variable contains the value '$env' it checks the
    '$env' key within the variables dict for the corresponding
    dictionary key. If it is found, the corresponding value is
    returned. If not, a TricotEnvVariableError is raised.

    When the value of a variables is not '$env', it is directly
    returned.

    Parameters:
        variables       Variables dictionary created during the Tester initialization
        key             Key of the current variable in question
        value           Value of the current variable in question

    Returns:
        value       Value with possible runtime variables applied
    '''
    if value != '$env':
        return value

    try:
        return (variables['$env'])[key]

    except KeyError:
        raise TricotEnvVariableError(f"Unable to find environment variable '{key}'.")


def apply_variables(candidate: Any, var_dict: dict[str, Any] = None) -> Any:
    '''
    Applies the variables contained in the Plugin on each str value within the
    parameters. For dictionaires and list, the function uses recursion to iterate
    over all possible items.

    Parameters:
        candidate       Current param value. Required for recursion.
        var_dict        Variables to apply.

    Returns:
        None
    '''
    if var_dict is None:
        return candidate

    cur_type = type(candidate)

    if cur_type is dict:
        for key, value in candidate.items():
            candidate[key] = apply_variables(value, var_dict)

    elif cur_type is list:
        for ctr in range(len(candidate)):
            candidate[ctr] = apply_variables(candidate[ctr], var_dict)

    elif cur_type is str:

        for var, var_value in var_dict.items():

            var_value = resolve_runtime_variables(var_dict, var, var_value)
            var_value = resolve_env_variables(var_dict, var, var_value)
            variable_key = '${'+str(var)+'}'

            if candidate == variable_key:
                candidate = var_value

                if type(candidate) is not str:
                    break

            else:
                candidate = candidate.replace(variable_key, str(var_value))

    return candidate


def make_ordinal(n: int) -> str:
    '''
    Convert an integer into its ordinal representation. Used for pretty printing and copied
    from: https://stackoverflow.com/questions/9647202/ordinal-numbers-replacement

        make_ordinal(0)   => '0th'
        make_ordinal(3)   => '3rd'
        make_ordinal(122) => '122nd'
        make_ordinal(213) => '213th'
    '''
    n = int(n)
    suffix = ['th', 'st', 'nd', 'rd', 'th'][min(n % 10, 4)]
    if 11 <= (n % 100) <= 13:
        suffix = 'th'
    return str(n) + suffix


def check_keys(expected_keys: list[str], yaml_dict: dict) -> None:
    '''
    What happes quite some time is that validators are indented incorrectly and appear within the test
    section of the .yml definition. This is annoying, as your test seems to work, but was actually never
    validated. This function checks for unexpcted keys within .yml files and prints a warning if one is
    encountered.

    Parameters:
        expected_keys   List of keys that are expected to appear
        yaml_dict       Python dirctionary that represents the .yml file

    Returns:
        None
    '''
    for key in yaml_dict.keys():
        if key not in expected_keys:
            tricot.Logger.print_yellow('Warning:', end=' ')
            tricot.Logger.print_mixed_blue_plain('Test', yaml_dict['title'], 'contains unexpected key', end=': ')
            tricot.Logger.print_yellow_plain(key)


def merge_environment(new: dict, current: dict, path: Path) -> dict:
    '''
    Helper function to merge two environment dictionaries. Is mainly used to handle
    the case where the new environment is not a dictionary.

    Parameters:
        new             New environment variables to add
        current         Current set of environment variables
        path            Path of the currently parsed configuration file

    Returns:
        environment     Merged environment variables
    '''
    current = current or {}

    if new is None:
        return current

    if type(new) is not dict:
        raise tricot.TestKeyError(None, path, "Key 'env' needs to be a dictionary in the ")

    return {**new, **current}


def merge_default_environment(env: dict) -> dict:
    '''
    Merges the current user environment with the specifies environment dictionary.

    Parameters:
        new             dict to combine with the default environment

    Returns:
        environment     Merged environment variables
    '''
    return {**os.environ, **env}


def add_environment(variables: dict[str, Any]) -> None:
    '''
    Adds the current user environment to the available variables within the
    $env key.

    Parameters:
        variables       Variable dictionary

    Returns:
        None
    '''
    variables['$env'] = os.environ



DOCKER_API = 1
FILE_NOT_FOUND = 2
KEYBOARD_INTERRUPT = 3
PARSER_ERROR = 4
PLUGIN_ERROR = 5
PLUGIN_EXCEPTION = 6
VARIABLE_ERROR = 7
TESTER_KEY_ERROR = 8
TEST_KEY_ERROR = 8
TRICOT_EXCEPTION = 9
TRICOT_RUNTIME_ERROR = 10
UNKNOWN = 11
VALIDATION_EXCEPTION = 12
VALIDATOR_ERROR = 13
VALUE_ERROR = 14
RUNTIME_ERROR = 15
CONDITION_FORMAT_ERROR = 16

LAST_ERROR = 0



from __future__ import annotations

import timeit
import subprocess
from typing import Any
from pathlib import Path

import tricot


class TricotRuntimeError(Exception):
    '''
    Custom exception class for exceptions raised when running the external command.
    '''
    def __init__(self, exception: Exception) -> None:
        '''
        Custom exception class that stores the original exception within a variable.
        '''
        self.original = exception
        super().__init__('Runtime Error')


class Command:
    '''
    The command class represents a test command and is responsible for running
    the command and storing the corresponding outputs along with some meta
    information.
    '''
    def __init__(self, command: list) -> None:
        '''
        Crates a new Command object.

        Parameters:
            command     Command that is executed during a test.

        Returns:
            None
        '''
        self.path = None
        self.stdout = None
        self.stderr = None
        self.status = None
        self.runtime = None

        self.command = command

    def run(self, path: Path, timeout: int, hotplug_variables: dict[str, Any] = None, env: dict = {}):
        '''
        Just a wrapper around the actual command execution function. It is basically used
        to reduce the complexity of the run function and to handle the special case of
        the ${prev} variable within the command specification. It returns nothing, as
        outputs and meta information is stored within class variables.

        Parameters:
            path                File system path where the command is run in
            timeout             Timeout to use during command execution
            hotplug_variables   Variables that are applied at runtime.
            env                 Environment variables

        Returns:
            None
        '''
        if self.command[0] != '${prev}':
            self.command = tricot.Test.apply_variables(self.command, hotplug_variables)

            try:
                self.path = path
                timer = timeit.Timer(lambda: self._run(path, timeout, env))
                self.runtime = timer.timeit(number=1)

            except Exception as e:
                tricot.Logger.print_plain_red("error.")
                raise TricotRuntimeError(e)

            except KeyboardInterrupt as e:
                tricot.Logger.print_plain_red("canceled.")
                raise KeyboardInterrupt(e)

        else:
            prev = hotplug_variables.get('$prev')

            if prev is not None and prev.validate_run():
                self.copy(prev)

            else:
                tricot.Logger.print_plain_red("error.")
                raise tricot.TricotException("Special '${prev}' variable was used, but no previous output exists.")

    def _run(self, path: Path, timeout: int, env: dict = {}) -> None:
        '''
        Performs the actual command execution via subprocess.Popen. All relevant outputs and
        status codes are saved within of object attributes.

        Parameters:
            path                File system path where the command is run in
            timeout             Timeout to use during command execution
            env                 Environment variables

        Returns:
            cmd_output      Command output in the form [status_code, stdout & stderr]
        '''
        envi = tricot.utils.merge_default_environment(env)

        try:
            process = subprocess.Popen(self.command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=path, env=envi)
            self.stdout, self.stderr = process.communicate(timeout=timeout)
            self.status = process.returncode

        except subprocess.TimeoutExpired:
            process.kill()
            self.stdout, self.stderr = process.communicate()
            self.status = 99

        except subprocess.CalledProcessError as e:
            self.stdout = e.output
            self.stderr = e.stderr
            self.status = e.returncode

        self.stdout = self.stdout.decode('utf-8')
        self.stderr = self.stderr.decode('utf-8')

    def copy(self, other: Command) -> None:
        '''
        Copies all attribute values from the specified Command into the current object.

        Parameters:
            other       Command object to copy the values from

        Returns:
            None
        '''
        attrs = list(filter(lambda x: not x.startswith('__'), self.__dict__))
        for attr in attrs:
            setattr(self, attr, getattr(other, attr))

    def validate_run(self) -> bool:
        '''
        Validates that the command object was run. This is done by checking attributes that
        are set during command execution like self.stdout, self.stderr or self.status.

        Parameters:
            None

        Returns:
            boolean     True if command was run and contains all required attributes.
        '''
        attrs = list(filter(lambda x: not x.startswith('__'), self.__dict__))
        for attr in attrs:

            value = getattr(self, attr)
            if value is None:
                return False

        return True

    def get_output(self) -> str:
        '''
        Returns the merged stdout and stderr outputs. Outputs are separated with a single
        newline.

        Parameters:
            None

        Returns:
            output      Returns merged stdout and stderr output
        '''
        output = ''

        if self.stdout:
            output += self.stdout

        if self.stderr:
            output += '\n'
            output += self.stderr

        return output



from .command import *
from .condition import *
from .constants import *
from .docker import *
from .logging import *
from .plugin import *
from .template import *
from .tricot import *
from .validation import *

name = 'tricot'



from __future__ import annotations

import time
import docker
import atexit
from typing import Any
from pathlib import Path

import tricot


class TricotContainer:
    '''
    The TricotContainer class represents a docker container that was started by tricot.
    '''

    def __init__(self, name: str, image: str, env: list[str] = None, volumes: dict[str, str] = None,
                 aliases: dict[str, str] = {}) -> None:
        '''
        Initializes the container and registers an atexit event, but does not start the container.

        Parameters:
            name        Name of the running container
            image       Name of the image to run
            env         Environment variables to assign to the container
            volumes     Volumes to assign to the container
            aliases     Aliases for docker variables

        Returns:
            None
        '''
        self.name = name
        self.image = image

        self.env = env
        self.volumes = volumes
        self.aliases = aliases

        self.client = docker.from_env()
        self.container = None

        atexit.register(self.stop_container)

    def start_container(self) -> None:
        '''
        Starts the container and print somes general information about it. The start
        procedure currently includes a shot timeout of one second. When skipping this,
        it can be the case that the container is not already full up and running when
        other functions are called. As a result, docker variables like the IP address
        might be empty.

        Parameters:
            None

        Returns:
            None
        '''
        tricot.Logger.print('')
        tricot.Logger.print_mixed_yellow('Starting container:', self.name)
        tricot.Logger.increase_indent()

        tricot.Logger.print_mixed_blue('Image:', self.image)
        tricot.Logger.print_mixed_blue('Volumes:', str(self.volumes))
        tricot.Logger.print_mixed_blue('Environment:', str(self.env))

        tricot.Logger.decrease_indent()

        self.client.containers.run(self.image, name=self.name, volumes=self.volumes,
                                   environment=self.env, detach=True, auto_remove=True)

        self.container = self.client.containers.get(self.name)
        time.sleep(1)

    def stop_container(self) -> None:
        '''
        Stops the container. This function does not need to be called manually, as it
        is registered as an exit event during the object initialization.

        Parameters:
            None

        Returns:
            None
        '''
        if self.container:

            tricot.Logger.print('')
            tricot.Logger.print_mixed_yellow('Stopping container:', self.name)

            self.container.stop()
            self.container = None

    def get_container_variables(self) -> dict[str, str]:
        '''
        Returns a dictionary of container variables. Currently, only the IP address and
        the mounted volumes are included. Additional data may be added in future.

        Parameters:
            None

        Returns:
            dict        Container variables
        '''
        variables = {f'DOCKER-{self.name}-IP': self.container.attrs['NetworkSettings']['IPAddress']}
        variables[f'DOCKER-{self.name}-GATEWAY'] = self.container.attrs['NetworkSettings']['Gateway']

        ctr = 0
        for key, value in self.volumes.items():
            variables[f'DOCKER-{self.name}-VOLUME{ctr}'] = value['bind']
            variables[f'DOCKER-{self.name}-VOLUME{ctr}-HOST'] = key

        for key, value in self.aliases.items():

            try:
                variables[value] = variables[key]

            except KeyError:
                pass

        return variables

    def from_list(input_list: list, path: Path, variables: dict[str, Any]) -> list[TricotContainer]:
        '''
        Returns a list of TricotContainer that were created according to the input list.
        The input list is expected to be the result read in from the containers section
        of a .yml file.

        Parameters:
            input_list      Input list containing container specifications
            path            Path object to the yaml file where the container was defined in
            variables       Variables to apply.

        Returns:
            list            List of corresponding TricotContainer objects
        '''
        containers = list()

        try:

            for item in input_list:

                volume_dict = dict()
                volumes = item.get('volumes', [])
                volumes = tricot.utils.apply_variables(volumes, variables)

                for volume in volumes:
                    split = volume.split(':')

                    if '/' in split[0]:
                        p = Path(split[0])

                        if not p.is_absolute():
                            p = path.parent.joinpath(p).absolute()

                    if len(split) < 3:
                        split.append('rw')

                    volume_dict[str(p)] = {'bind': split[1], 'mode': split[2]}

                name = tricot.utils.apply_variables(item['name'])
                image = tricot.utils.apply_variables(item['image'])
                aliases = tricot.utils.apply_variables(item.get('aliases', {}))
                environment = tricot.utils.apply_variables(item.get('env', {}))

                container = TricotContainer(name, image, environment, volume_dict, aliases)
                containers.append(container)

            return containers

        except KeyError as e:
            raise tricot.TesterKeyError(str(e), section='containers')



from __future__ import annotations

import os
import sys
import time
import shutil
import atexit
import socket
import functools
import threading
import subprocess
import http.server
from typing import Any
from pathlib import Path

import tricot.utils


this = sys.modules[__name__]
this.plugins = {}


def register_plugin(plugin_name: str, plugin_class: type) -> None:
    '''
    Registers a plugin class under the specified name. This function needs to
    be called to make plugins available within of .yml files. Some default
    plugin classes are registerd at the end of this file. Others can be manually
    added by users that use tricot as a library.

    Parameters:
        plugin_name         Name of the plugin to use in .yml files
        plugin_class        Reference to the corresponding plugin class

    Returns:
        None
    '''
    this.plugins[plugin_name] = plugin_class


def get_plugin(path: Path, plugin_name: str, param: Any, variables: dict[str:str]) -> Plugin:
    '''
    Searches for the specified plugin name within the registred plugins and
    uses the corresponding class to create an instance for the plugin. This instance
    is returned by the function. If the specified plugin name is not found, the
    function raises a PluginError.

    Parameters:
        path                Path object to the .yml file containing the plugin
        plugin_name         Name of the requested plugin
        param               Params to initialize the plugin with
        variables           Variables to initialize the plugin with
    '''
    plug_class = this.plugins.get(plugin_name)

    if plug_class is None:
        raise PluginError(path, f"Unable to find specified plugin '{plugin_name}'.")

    return plug_class(path, plugin_name, param, variables)


def get_plugin_list() -> list[str]:
    '''
    Returns a list of currently registered plugin names.

    Parameters:
        None

    Returns:
        plugin_list       List of registred plugins
    '''
    keys = this.plugins.keys()
    return list(keys)


class PluginException(Exception):
    '''
    PluginExceptions are raised by plugins when they throw any other kind of exception.
    They contain the original exception as parameter.
    '''
    def __init__(self, exception: Exception, name: str, path: Path) -> None:
        '''
        Custom exception class that stores the original exception within a variable.
        '''
        self.original = exception
        self.name = name
        self.path = path

        super().__init__('PluginException')


class PluginError(Exception):
    '''
    PluginErrors are raised when there are problems with initializing a plugin.
    '''
    def __init__(self, path: Path, message: str) -> None:
        '''
        Adds the additional path attribute (location of the corresponding configuration file)
        '''
        self.path = str(path.resolve())
        super().__init__(message)


class Plugin:
    '''
    The Plugin class is basically an abstract class with the only purpose of
    being extended by other classes. All possible plugins implemented in
    tricot are based on this superclass that contains some shared methods and
    some that need to be overwritten by the child classes.

    Within test.yaml files, Plugins can be defined in different ways:

        * example-plugin: This is a value
        * example-plugin2:
            param1: value1
            param2: value2
            ...
        * example-plugin3:
            - value1
            - value2
            ...

    Which form to choose depends on the corresponding Plugin requirements. When
    choosing the first version, the 'param' attribute of the Plugin just contains
    the value specified in the yaml file. When choosing the second or third form, the
    'param' attribute contains a dictionary or list with the specified parameters.

    Plugins can use the 'param_type' class variable to indicate which type of parameters
    they expect. For collection types like dict or list, it is also possible to specify
    the 'inner_types' attribute to specify requirements on the containing types.
    To understand this feature you probably want to look at the HttpListenerPlugin in
    this file, which uses the 'inner_types' variable.

    Custom Plugins should at least overwrite the 'run' method to perform their
    operation. Other methods can be overwritten or added on demand. The run
    should perform its operation without returning anything. PluginExceptions do not
    need to be raised manually. Invalid configuration within the .yml file should cause
    a PluginError, but it is recommended to only raise this error during initialization
    and not within the 'run' method.
    '''
    param_type = None
    inner_types = None

    def __init__(self, path: Path, name: str, param: Any, variables: dict[str, Any]) -> None:
        '''
        Initializes the Plugin.

        Parameters:
            path        Path object to the .yml file containing the plugin
            name        Name of the Plugin (as specified in the .yml file)
            param       Parameters of the Plugin (can be str, dict, int, bool, ...)
            variables   Variables contained in this dict are replaced in all str parameters
        '''
        self.path = path
        self.name = name
        self.param = param
        self.variables = variables
        self.stopped = True
        self.check_param_type()

    def check_param_type(self) -> None:
        '''
        Checks whether the specified parameter type matches the expected one. If this is not
        the case, a PluginError is raised. As parameter types may changed after variable
        substituation, this function also applies variables before checking.

        Parameters:
            None

        Returns:
            None
        '''
        self.param = tricot.utils.apply_variables(self.param, self.variables)

        if self.param_type is None:
            return

        if type(self.param) is not self.param_type:
            raise PluginError(self.path, f"Plugin '{self.name}' requires a parameter type of {str(self.param_type)}")

        self.check_inner_types()

    def check_inner_types(self) -> None:
        '''
        Plugins that require a dictionary or list can specify the static class variable 'inner_types'
        to specify which inner types are expecetd within the collection types. When a list is used,
        'inner_types' is expected to contain a list with allowed types. In case of a dict, 'inner_types'
        is expected to contain a dict, containing the possible parameters as keys and a dict of
        {'required': bool, 'type': type} as value.

        The check is currently only implemented on the first layer (non recursive).

        Parameters:
            None

        Returns:
            None
        '''
        if type(self.param) is dict and type(self.inner_types) is dict:

            for key, value in self.inner_types.items():

                param = self.param.get(key)
                if value['required'] and not param:
                    message = f"Plugin '{self.name}' requires key with name '{key}' and type {value['type']}."
                    raise PluginError(self.path, message)

                if param is not None and type(param) is not value['type']:
                    message = f"Plugin '{self.name}' expects type {value['type']} for the '{key}' key."
                    raise PluginError(self.path, message)

        elif type(self.param) is list and type(self.inner_types) is list:

            for param in self.param:
                if type(param) not in self.inner_types:
                    message = f"Plugin '{self.name}' requires a parameter type of list[{str(self.inner_types)}."
                    raise PluginError(self.path, message)

    def from_list(path: Path, input_list: list[dict], variables: dict[str, Any] = {}) -> list[Plugin]:
        '''
        Constructs a new list of Plugins from a python list. The list content is usually
        the specified plugin configuration within a .yml file.

        Parameters:
            path        Path object to the .yml file containing the plugin
            input_list  Python list as read in from a .yml file
            variables   Global variables that should be inherited by the plugin

        Returns:
            Plugin   Newly constructed Plugin objects
        '''
        if input_list is None:
            return []

        if type(input_list) is dict:
            raise PluginError(path, "Plugins need to be specified as a list.")

        plugins = list()

        for item in input_list:
            for key, value in item.items():
                val = get_plugin(path, key, value, variables)
                plugins.append(val)

        return plugins

    def _run(self, hotplug_variables: dict[str, Any] = None) -> None:
        '''
        This method is called internally to perform the Plugin action. It is basically
        a wrapper to the user defined 'run' method that catches possible exceptions and repalces
        them with the PluginException class.

        Parameters:
            hotplug_variables   Variables to apply at runtime

        Returns:
            None
        '''
        self.stopped = False
        atexit.register(self._stop)
        self.param = tricot.utils.apply_variables(self.param, hotplug_variables)

        try:
            self.run()

        except Exception as e:
            raise PluginException(e, self.name, self.path)

    def run(self) -> None:
        '''
        Dummy run method. This one needs to be overwritten by the actual Plugin implementations.
        '''
        return None

    def _stop(self) -> None:
        '''
        This method is called internally to perform Plugin cleanup actions. It is run
        after all Tests and Testers from the .yml file where the Plugin was defined in
        have finished. It is basically a wrapper around the user defined 'stop' method
        that performs the actual Plugin cleanup.

        Whether or not a plugin requires a cleanup depends on the plugins functionality.
        Therefore, this method does not need to be implemented.

        Parameters:
            None

        Returns:
            None
        '''
        if self.stopped:
            return

        try:
            self.stop()
            self.stopped = True

        except Exception as e:
            atexit.unregister(self._stop)
            raise PluginException(e, self.name, self.path)

    def stop(self) -> None:
        '''
        Dummy stop method. If a Plugin requires cleanup, this method should be overwritten.
        '''
        return None

    def resolve_path(self, path: str) -> str:
        '''
        Resolves the specified path relative to the current plugin definition (file location of
        the plugins .yml file). If the input path is already absolute, it is just returned.

        Parameters:
            path        File system path to resolve

        Returns:
            resolved    Resolved file system path
        '''
        if os.path.isabs(path):
            return Path(path)

        return self.path.parent.joinpath(path)


class OsCommandPlugin(Plugin):
    '''
    The OsCommandPlugin simply executes the specified operating system command.
    Commands are executed from the directory where the plugin was configured in.

    Example:

        plugins:
            - os_command:
                init: 1
                cmd:
                    - nc
                    - -vlp
                    - 4444
    '''
    param_type = dict
    inner_types = {
                    'ignore_error': {'required': False, 'type': bool},
                    'init': {'required': False, 'type': int},
                    'background': {'required': False, 'type': bool},
                    'timeout': {'required': False, 'type': int},
                    'cmd': {'required': True, 'type': list}
                  }

    def on_exit(self, command) -> None:
        '''
        Just a helper function that is called when a process exited.
        '''
        ignore_error = self.param.get('ignore_error', False)
        poll = self.process.poll()

        if poll and poll != 0 and not ignore_error:
            command = ' '.join(command)
            raise OSError(f"Command '{command}' exited with a non zero status code.")

    def run(self) -> None:
        '''
        Run the specified operating system command.
        '''
        init = self.param.get('init', 0)

        command = self.param['cmd']
        timeout = self.param.get('timeout', 0)
        background = self.param.get('background', False)

        for ctr in range(len(command)):
            command[ctr] = str(command[ctr])

        self.process = subprocess.Popen(command, cwd=self.path.parent, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        if timeout > 0:
            self.process.communicate(timeout=timeout)

        elif not background:
            self.process.wait()

        if init > 0:
            time.sleep(init)

        self.on_exit(command)

    def stop(self) -> None:
        '''
        Stop the process if still existing.
        '''
        if hasattr(self, 'process') and self.process and self.process.poll() is None:
            self.process.kill()


class MkdirPlugin(Plugin):
    '''
    The MkdirPlugin simply creates the specified directories.
    Relative paths are interpreted to the directory the plugin was configured in.

    Example:

        plugins:
            - mkdir:
                cleanup: true
                force: false
                dirs:
                    - /tmp/test
                    - /tmp/test2
    '''
    param_type = dict
    inner_types = {
                    'cleanup': {'required': False, 'type': bool},
                    'force': {'required': False, 'type': bool},
                    'dirs': {'required': True, 'type': list}
                  }

    blacklist = ['/', '/home', '/opt', '/var']
    directories = list()

    def run(self) -> None:
        '''
        Create the specified directories if not already existing.
        '''
        for directory in self.param['dirs']:

            directory = self.resolve_path(directory)
            os.makedirs(directory, exist_ok=True)

            if directory not in MkdirPlugin.directories:
                MkdirPlugin.directories.append(directory)

                if self.directories is None:
                    self.directories = []

                self.directories.append(directory)

    def stop(self) -> None:
        '''
        Remove all directories created by the plugin (only if cleanup is set to true).
        '''
        if not self.param.get('cleanup', False) or not hasattr(self, 'directories'):
            return

        for directory in self.directories:
            if directory not in MkdirPlugin.blacklist:

                try:
                    os.rmdir(directory)

                except OSError as e:

                    if 'not empty' in str(e) and self.param.get('force', False):
                        shutil.rmtree(directory)

                MkdirPlugin.directories.remove(directory)
                self.directories.remove(directory)


class HttpListenerPlugin(Plugin):
    '''
    The HttpListenerPlugin starts a HTTP server in the background and servers files
    from the specified directory and port.

    Example:

        plugins:
            - http_listener:
                port: 8000
                dir: ./www
    '''
    param_type = dict
    inner_types = {
                    'port': {'required': True, 'type': int},
                    'dir': {'required': True, 'type': str}
                  }
    instances = []

    def __init__(self, *args, **kwargs) -> None:
        '''
        Make sure that the specified port is valid before running the server.
        '''
        super().__init__(*args, **kwargs)

        if self.param['port'] <= 0 or self.param['port'] > 65535:
            raise PluginError(self.path, f"Specified port '{self.param['port']}' is invalid. Needs to be between 0-65535.")

    class CustomHandler(http.server.SimpleHTTPRequestHandler):
        '''
        Custom HTTPRequestHandler that uses the base functioanlity of SimpleHTTPRequestHandler
        and only overwrites its log_message method to prevent any otuputs.
        '''

        def log_message(self, format, *args):
            '''
            Overwrite log_message to prevent log messages from being printed.
            '''
            return

    def run(self) -> None:
        '''
        Checks whether a server on this port is already running and starts it otherwise.
        '''
        port = self.param['port']
        directory = self.param['dir']

        if port in HttpListenerPlugin.instances:
            return

        directory = self.resolve_path(directory)
        if not directory.is_dir():
            raise FileNotFoundError(f"Specified directory '{directory.absolute()}' does not exist.")

        self.thread = threading.Thread(name=f'http-{port}', target=self.start_server, args=(port, directory))
        self.thread.setDaemon(True)
        self.thread.start()

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        while sock.connect_ex(('127.0.0.1', port)) != 0:
            time.sleep(0.1)
        sock.close()

        HttpListenerPlugin.instances.append(port)

    def start_server(self, port: int, directory: str) -> None:
        '''
        Starts the HTTPListener. Should be run in a separate thread.
        '''
        handler = functools.partial(HttpListenerPlugin.CustomHandler, directory=directory)
        self.server = http.server.HTTPServer(('', port), handler)
        self.server.serve_forever()

    def stop(self) -> None:
        '''
        Stops the HTTPListener and the corresponding thread.
        '''
        if hasattr(self, 'server') and self.server:
            self.server.server_close()
            self.server.shutdown()

        if hasattr(self, 'thread') and self.thread:
            self.thread.join()

        if self.param['port'] in HttpListenerPlugin.instances:
            HttpListenerPlugin.instances.remove(self.param['port'])


class CleanupPlugin(Plugin):
    '''
    The CleanupPlugin is used to remove files and directories that have been created during
    tests. Relative paths are resolved to the directory the plugin is defined in.

    Example:

        plugins:
            - cleanup:
                force: false
                items:
                    - /tmp/test
                    - /tmp/test2
    '''
    param_type = dict
    inner_types = {
                    'force': {'required': False, 'type': bool},
                    'items': {'required': True, 'type': list}
                  }

    blacklist = ['/', '/home', '/opt', '/var']

    def stop(self) -> None:
        '''
        Remove all items specified in the plugin definition.
        '''
        for item in self.param.get('items', []):

            item = self.resolve_path(item)
            try:

                if item in CleanupPlugin.blacklist:
                    continue

                elif os.path.isdir(item):

                    try:
                        os.rmdir(item)

                    except OSError as e:
                        if 'not empty' in str(e) and self.param.get('force', False):
                            shutil.rmtree(item)

                else:
                    os.remove(item)

            except FileNotFoundError:
                continue


class CleanupCommandPlugin(Plugin):
    '''
    Basically a copy of the OsCommandPlugin, but runs on the end of the corresponding tester.

    Example:

        plugins:
            - cleanup_command:
                ignore_error: True
                cmd:
                    - rm
                    - -r
                    - /tmp/testdir
    '''
    param_type = dict
    inner_types = {
                    'ignore_error': {'required': False, 'type': bool},
                    'timeout': {'required': False, 'type': int},
                    'cmd': {'required': True, 'type': list}
                  }

    def on_exit(self, command) -> None:
        '''
        Just a helper function that is called when a process exited.
        '''
        ignore_error = self.param.get('ignore_error', False)
        poll = self.process.poll()

        if poll and poll != 0 and not ignore_error:
            command = ' '.join(command)
            raise OSError(f"Command '{command}' exited with a non zero status code.")

    def stop(self) -> None:
        '''
        Run the specified command on stop.
        '''
        command = self.param['cmd']
        timeout = self.param.get('timeout', 0)

        for ctr in range(len(command)):
            command[ctr] = str(command[ctr])

        self.process = subprocess.Popen(command, cwd=self.path.parent, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        if timeout > 0:
            self.process.communicate(timeout=timeout)

        else:
            self.process.wait()

        self.on_exit(command)


class CopyPlugin(Plugin):
    '''
    The CopyPlugin can be used to copy files before a test runs. It also offers a cleanup
    action to remove copied files after the test.

    Example:

        plugins:
            - copy:
                cleanup: True
                recursive: False
                from:
                    - /tmp/from-here
                to:
                    - /opt/to-there
    '''
    param_type = dict
    inner_types = {
                    'cleanup': {'required': False, 'type': bool},
                    'from': {'required': True, 'type': list},
                    'to': {'required': True, 'type': list}
                  }

    blacklist = ['/', '/home', '/opt', '/var']

    def run(self) -> None:
        '''
        Copy the specified files into the target location.
        '''
        self.cleanup = []

        dest = self.param.get('to', [])
        files = self.param.get('from', [])

        if len(dest) != len(files):
            raise ValueError("The 'from' and 'to' parameters need to be equally sized lists.")

        for ctr in range(len(files)):

            src_path = self.resolve_path(files[ctr])
            dest_path = self.resolve_path(dest[ctr])

            if src_path.is_file():
                created = shutil.copy(src_path, dest_path)
                self.cleanup.append(Path(created))

            elif src_path.is_dir():

                if dest_path.is_dir():
                    dest_path = dest_path.joinpath(src_path.name)

                created = shutil.copytree(src_path, dest_path)
                self.cleanup.append(Path(created))

    def stop(self) -> None:
        '''
        Removed copied files if desired.
        '''
        if not self.param.get('cleanup', False):
            return

        for item in self.cleanup:

            if not item.exists():
                continue

            elif item.absolute() in CopyPlugin.blacklist:
                raise ValueError("Plugin attempted to remove the blacklisted ressource '{item}'.")

            elif item.is_file():
                item.unlink()

            elif item.is_dir():
                shutil.rmtree(str(item.absolute()))


register_plugin("os_command", OsCommandPlugin)
register_plugin("mkdir", MkdirPlugin)
register_plugin("http_listener", HttpListenerPlugin)
register_plugin("cleanup", CleanupPlugin)
register_plugin("cleanup_command", CleanupCommandPlugin)
register_plugin("copy", CopyPlugin)



#!/usr/bin/python3

import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    url='https://github.com/qtc-de/tricot',
    name='tricot',
    author='Tobias Neitzel (@qtc_de)',
    version='1.3.1',
    author_email='',

    description='Trivial Command Testser',
    long_description=long_description,
    long_description_content_type='text/markdown',

    packages=['tricot'],
    install_requires=[
                        'PyYAML',
                        'termcolor',
                        'docker'
                     ],
    scripts=[
                'bin/tricot',
            ],
    classifiers=[
                    'Programming Language :: Python :: 3',
                    'Operating System :: Unix',
                    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
                ],
)
